import numpy as np
import matplotlib.pyplot as plt
from opensimplex import OpenSimplex


class UniformPerlinMap:

    def __init__(self, x_dim, y_dim, scale, sheet_x=0, sheet_y=0, rnd_seed=1):
        self.x_dim = x_dim
        self.y_dim = y_dim
        self.scale = scale
        self.sheet_x = sheet_x
        self.sheet_y = sheet_y
        self.rnd_seed = rnd_seed

        noise = OpenSimplex(seed=self.rnd_seed)
        nrow = self.x_dim
        size = self.sheet_x

        x = y = np.linspace(0, size, nrow)
        n = [[noise.noise2d(i / scale, j / scale) for j in y] for i in x]
        m = np.concatenate(n)
        sorted_idx = np.argsort(m)
        max_val = nrow * 2
        idx = len(m) // max_val
        for ii, val in enumerate(range(max_val)):
            m[sorted_idx[ii * idx:(ii + 1) * idx]] = val
        p_map = (m - nrow) / nrow
        self.map = p_map.reshape(nrow, -1)
        self.map *= np.pi

    def get_meshgrid_of_neuron_positions(self):

        xmin = np.min(0.)
        xmax = np.max(self.sheet_x)
        dx = (xmax - xmin) / (self.x_dim - 1)

        ymin = np.min(0.)
        ymax = np.max(self.sheet_y)
        dy = (ymax - ymin) / (self.y_dim - 1)

        X, Y = np.meshgrid(np.arange(xmin, xmax + 2 * dx, dx) - dx / 2., np.arange(ymin, ymax + 2 * dy, dy) - dy / 2.)

        return X, Y

    def get_tuning(self, x, y):
        id_x = int(x * (self.x_dim - 1) / self.sheet_x)
        id_y = int(y * (self.y_dim - 1) / self.sheet_y)
        return self.map[id_x, id_y]

    def get_tuning_by_id(self, id_x, id_y):
        return self.map[id_x, id_y]

    def plot_map(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1, 1)

        X, Y = self.get_meshgrid_of_neuron_positions()

        Z = np.zeros((self.x_dim, self.y_dim))

        for y_idx in range(Z.shape[1]):
            for x_idx in range(Z.shape[0]):
                o_map_val = self.get_tuning_by_id(x_idx, y_idx)
                Z[x_idx, y_idx] = o_map_val
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        plt.set_cmap('twilight')
        pcm = ax.pcolormesh(X, Y, Z.T)
        plt.gcf().colorbar(pcm, ax=ax)
